# copyright (C) 1997-1999 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: print.tcl,v 1.6.1.6 2000/01/02 18:45:24 jfontain Exp $}


namespace eval print {

    variable dotsPerMillimeter [expr {72/25.4}]                                          ;# 72 dots/inch is the default for printers
    variable helpMessage \
{Print to a printer device or a file in Postscript.
When printing to a file, you may choose it with the file browser.
Note that this file is also used and overwritten for previewing,
which uses the gv (ghostview) utility.
Also use the preferences dialog box for setting printing defaults.}

    proc printOrSaveCanvas {} {
        variable printToFile $global::printToFile
        variable printCommand $global::printCommand
        variable fileToPrintTo $global::fileToPrintTo
        variable printOrientation $global::printOrientation
        variable printPalette $global::printPalette
        variable printPaperSize $global::printPaperSize
        variable helpMessage

        set objects {}                                                                  ;# to delete upon destruction of this folder

        set dialog\
            [new dialogBox . -buttons oc -default o -title {moodss: Print} -die 0 -x [winfo pointerx .] -y [winfo pointery .]]
        set toplevel $widget::($dialog,path)
        set frame [frame $toplevel.frame]

        set row 0
        if {[catch {$global::canvas postscript -colormode mono -width 1 -height 1 -prolog 0}]} {
            message $frame.warning -width [winfo screenwidth .] -font $font::(mediumBold) -justify left -text \
{WARNING: your Tcl/Tk interpreter does not seem to include the dash patch
(ftp://ftp.neosoft.com/pub/tcl/sorted/packages-8.0/devel/tk8.0.5dash.patch.gz)
and you are likely to obtain a blank page as printout.}
            grid $frame.warning -pady 5 -row $row -column 0 -columnspan 3
            incr row
        }
        message $frame.help -width [winfo screenwidth .] -font $font::(mediumNormal) -justify left -text $helpMessage
        grid $frame.help -pady 5 -row $row -column 0 -columnspan 3

        incr row
        radiobutton $frame.toCommand -variable print::printToFile -value 0 -text {with Command:}
        grid $frame.toCommand -row $row -column 0 -sticky w
        entry $frame.command -textvariable print::printCommand
        grid $frame.command -row $row -column 1 -sticky ew
        button $frame.preview -text Preview... -command "wm withdraw $toplevel; print::preview; wm deiconify $toplevel"
        grid $frame.preview -row $row -column 2 -sticky ew
        incr row
        radiobutton $frame.toFile -variable print::printToFile -value 1 -text {or to File:}
        grid $frame.toFile -row $row -column 0 -sticky w
        entry $frame.file -textvariable print::fileToPrintTo
        grid $frame.file -row $row -column 1 -sticky ew
        button $frame.browse -text Browse... -command "print::inquireFileToPrintTo $frame"
        grid $frame.browse -row $row -column 2 -sticky ew
        if {$printToFile} {
            $frame.toFile invoke
        } else {
            $frame.toCommand invoke
        }

        incr row
        grid [label $frame.orientation -text Orientation:] -row $row -column 0 -sticky w
        set entry [new comboEntry $frame -font $widget::(default,EntryFont) -list $global::printOrientations -editable 0]
        lappend objects $entry
        composite::configure $entry entry -textvariable print::printOrientation
        composite::configure $entry button -listheight [llength $global::printOrientations]
        composite::configure $entry button scroll -selectmode single
        grid $widget::($entry,path) -row $row -column 1 -columnspan 2 -sticky ew

        incr row
        grid [label $frame.palette -text Palette:] -row $row -column 0 -sticky w
        set entry [new comboEntry $frame -font $widget::(default,EntryFont) -list $global::printPalettes -editable 0]
        lappend objects $entry
        composite::configure $entry entry -textvariable print::printPalette
        composite::configure $entry button -listheight [llength $global::printPalettes]
        composite::configure $entry button scroll -selectmode single
        grid $widget::($entry,path) -row $row -column 1 -columnspan 2 -sticky ew

        incr row
        grid [label $frame.size -text {Paper size:}] -row $row -column 0 -sticky w
        set entry [new comboEntry $frame -font $widget::(default,EntryFont) -list $global::printPaperSizes -editable 0]
        lappend objects $entry
        composite::configure $entry entry -textvariable print::printPaperSize
        composite::configure $entry button -listheight [llength $global::printPaperSizes]
        composite::configure $entry button scroll -selectmode single
        grid $widget::($entry,path) -row $row -column 1 -columnspan 2 -sticky ew

        dialogBox::display $dialog $frame
        # make sure dialog box does not appear in printout
        widget::configure $dialog -command "delete $dialog; print::updateGlobals; update; print::print"

        bind $frame <Destroy> [concat delete $objects]                               ;# delete objects not managed by the dialog box
    }

    proc inquireFileToPrintTo {parentPath} {
        set file [tk_getSaveFile\
            -title {moodss: Print to file} -parent $parentPath -initialdir [pwd] -defaultextension .ps\
            -filetypes {{Postscript .ps} {{All files} *}} -initialfile $print::fileToPrintTo\
        ]
        if {[string length $file]>0} {                                                                              ;# not cancelled
            set print::fileToPrintTo $file
        }
    }

    proc updateGlobals {} {
        variable printToFile
        variable printCommand
        variable fileToPrintTo
        variable printOrientation
        variable printPalette
        variable printPaperSize

        set global::printToFile $printToFile
        set global::printCommand $printCommand
        set global::fileToPrintTo $fileToPrintTo
        set global::printOrientation $printOrientation
        set global::printPalette $printPalette
        set global::printPaperSize $printPaperSize
    }

    proc postscriptOptions {} {
        variable printOrientation
        variable printPaperSize
        variable printPalette
        variable dotsPerMillimeter

        update                                          ;# make sure everything is redrawn properly in case dialog box hid something

        set canvas $global::canvas
        # first calculate visible area coordinates
        foreach {left top right bottom} [$canvas cget -scrollregion] {}
        set width [expr {$right-$left}]
        set height [expr {$bottom-$top}]
        foreach {minimum maximum} [$canvas xview] {}
        set left [expr {$left+($minimum*$width)}]
        foreach {minimum maximum} [$canvas yview] {}
        set top [expr {$top+($minimum*$height)}]
        scan [winfo geometry $canvas] %ux%u width height                                                ;# now get visible area size
        set right [expr {$left+$width}]
        set bottom [expr {$top+$height}]
        # the eventually minimize to visible items
        foreach {boundsLeft boundsTop boundsRight boundsBottom} [$canvas bbox all] {}
        if {$boundsLeft>$left} {set left $boundsLeft}
        if {$boundsRight<$right} {set right $boundsRight}
        if {$boundsTop>$top} {set top $boundsTop}
        if {$boundsBottom<$bottom} {set bottom $boundsBottom}
        set width [expr {$right-$left}]
        set height [expr {$bottom-$top}]
        set inch 25.4                                                                                                 ;# millimeters
        set margin [expr {$inch/2}]                                                                 ;# all dimensions in millimeters
        switch -glob $printPaperSize {
            A3* {
                set pageWidth 297
                set pageHeight 420
            }
            A4* {
                set pageWidth 210
                set pageHeight 297
            }
            executive* {
                set pageWidth [expr {7.5*$inch}]
                set pageHeight [expr {10*$inch}]
            }
            legal* {
                set pageWidth [expr {8.5*$inch}]
                set pageHeight [expr {14*$inch}]
            }
            default {
                set pageWidth [expr {8.5*$inch}]                                                                           ;# letter
                set pageHeight [expr {11*$inch}]
            }
        }
        set pageX ${margin}m
        set rotate [expr {[string compare $printOrientation landscape]==0}]
        if {$rotate} {
            set pageY ${margin}m
        } else {
            set pageY [expr {$pageHeight-$margin}]m
        }
        # now make sure everything fits in page
        if {$rotate} {                                                                                 ;# swap page width and height
            set value $pageWidth
            set pageWidth $pageHeight
            set pageHeight $value
            unset value
        }
        set printWidth [expr {($pageWidth-(2*$margin))*$dotsPerMillimeter}]                                             ;# in pixels
        set printHeight [expr {($pageHeight-(2*$margin))*$dotsPerMillimeter}]
        set ratio 1                                                                  ;# use identical horizontal and vertical ratios
        if {$printWidth<$width} {
            set ratio [expr {$printWidth/$width}]
        }
        if {($printHeight<$height)&&(($printHeight/$height)<$ratio)} {
            set ratio [expr {$printHeight/$height}]
        }
        set options [list\
            -colormode $printPalette -rotate $rotate -x $left -y $top -width $width -height $height\
            -pageanchor nw -pagex $pageX -pagey $pageY\
        ]
        if {$ratio<1} {                                                                                     ;# size reduction needed
            lappend options -pagewidth [expr {$ratio*$width}] -pageheight [expr {$ratio*$height}]
        }
        return $options
    }

    proc print {} {
        variable printToFile
        variable fileToPrintTo
        variable printCommand

        set options [postscriptOptions]                       ;# do first as internal updating is done for proper geometry assertion
        if {$printToFile} {
            lifoLabel::push $global::messenger "printing to file $fileToPrintTo..."
        } else {
            lifoLabel::push $global::messenger printing...
        }
        update idletasks                                                                             ;# make sure message is visible
        if {$printToFile} {
            lappend options -file $fileToPrintTo
            eval $global::canvas postscript $options
        } else {                             ;# gather data first in case window gets obscured by printing utility (which may be gv)
            set data [eval $global::canvas postscript $options]
            set channel [open |$printCommand w]
            puts -nonewline $channel $data
            close $channel
        }
        lifoLabel::pop $global::messenger
    }

    proc preview {} {
        variable fileToPrintTo
        variable printPaperSize

        set options [postscriptOptions]                       ;# do first as internal updating is done for proper geometry assertion
        array set option $options
        set gvOptions -nodsc                                                                         ;# ignore comments to go faster
        if {$option(-rotate)} {                                                       ;# required for gv to precisely displqy output
            lappend gvOptions -landscape
        } else {
            lappend gvOptions -portrait
        }
        scan $printPaperSize %s media                                           ;# paper size format is: media (width x height unit)
        lappend gvOptions -media $media $fileToPrintTo
        lifoLabel::push $global::messenger "previewing with gv on file $fileToPrintTo..."
        update idletasks                                                                             ;# make sure message is visible
        lappend options -file $fileToPrintTo
        eval $global::canvas postscript $options
        eval blt::bgexec dummy gv $gvOptions &
        lifoLabel::pop $global::messenger
    }

}
